/* ----------------------------------------------------------------------------
   Ping Utility Library
   Copyright (C) 1994-2000 PING Software Group and Mark A Lindner

   This file is part of pingutil.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   ----------------------------------------------------------------------------
   $Log: dstring.c,v $
   Revision 1.3  2001/01/21 08:51:12  markl
   Fixed typo in function name.

   Revision 1.2  2000/07/04 12:54:35  markl
   More fixes.

   Revision 1.1  2000/06/14 06:02:43  ping
   Initial checkin.

   ----------------------------------------------------------------------------
*/

/* Feature test switches */

#include "config.h"

/* Local headers */

#include "pingutil.h"

/* Functions */

p_dstring_t *P_dstring_create(uint_t blocksz)
  {
  p_dstring_t *d;

  if(blocksz < P_DSTRING_MINBLOCKSZ)
    return(NULL);

  d = P_new(p_dstring_t);
  d->mem = P_newstr(blocksz);
  d->p = d->len = 0;
  d->blk = 1;
  d->blocksz = blocksz;
  return(d);
  }

/*
 */

char *P_dstring_destroy(p_dstring_t *d)
  {
  char *r;

  if(!d)
    return(NULL);

  r = d->mem;
  *(r + d->len) = NUL;
  r = P_realloc(r, d->len + 1, char);
  P_free(d);
  return(r);
  }

/*
 */

static char *__P_dstring_resize(p_dstring_t *d, uint_t rlen)
  {
  uint_t r;

  if(!d)
    return(NULL);

  r = d->p;
  if(((d->p += rlen) >= (d->blocksz * d->blk)) || !rlen)
    {
    d->blk = (d->p / d->blocksz) + 1;
    d->mem = P_realloc(d->mem, d->blk * d->blocksz, char);
    }
  if(!rlen || d->p > d->len)
    d->len = d->p;

  return(rlen ? (char *)(d->mem + r) : NULL);
  }

/*
 */

bool_t P_dstring_putc(p_dstring_t *d, char c)
  {
  char *r;

  if(!d || !c)
    return(FALSE);

  r = __P_dstring_resize(d, 1);
  *r = c;

  return(TRUE);
  }

/*
 */

bool_t P_dstring_puts(p_dstring_t *d, const char *s)
  {
  char *r;
  size_t l;

  if(!d || !s)
    return(FALSE);
  if(!*s)
    return(FALSE);

  l = strlen(s);
  r = __P_dstring_resize(d, (uint_t)l);
  memcpy((void *)r, (void *)s, l);

  return(TRUE);
  }

/*
 */

char P_dstring_getc(p_dstring_t *d)
  {

  if(!d)
    return (NUL);
  if(d->p >= d->len)
    return(NUL);

  return(*(d->mem + d->p++));
  }

/*
 */

char *P_dstring_gets(p_dstring_t *d, char *s, size_t len, char termin)
  {
  char *p, *t = s;
  size_t l = len;
  uint_t i;

  if(!d || !s || !len || !termin)
    return(NULL);
  if(d->p == d->len)
    return(NULL);

  for(p = d->mem + d->p, i = d->len - d->p; i-- && --l;)
    {
    if(*p == termin)
      break;
    *(t++) = *(p++);
    }
  *t = NUL;
  return(s);
  }

/*
 */

bool_t P_dstring_seek(p_dstring_t *d, unsigned long where, int whence)
  {
  unsigned long o;

  if(!d || !where) return(FALSE);
  switch(whence)
    {
    case P_DSTRING_SEEK_REL:
      o = d->p + where;    
      break;

    case P_DSTRING_SEEK_ABS:
      o = where;
      break;

    case P_DSTRING_SEEK_END:
      o = d->len - where;
      break;

    default:
      return(FALSE);
    }

  if(o < 0 || o > d->len)
    return(FALSE);
  d->p = o;

  return(TRUE);
  }

/*
 */

bool_t P_p_dstring_trunc(p_dstring_t *d, unsigned long size)
  {

  if(!d)
    return(FALSE);
  if(size > d->len)
    return(FALSE);

  d->p = size;
  __P_dstring_resize(d, 0);

  return(TRUE);
  }

/*
 */

p_dstring_t *P_dstring_load(const char *path, uint_t blocksz)
  {
  p_dstring_t *d;
  FILE *fp;
  struct stat st;
  uint_t i;
  char *p;
  size_t b;

  if(blocksz < P_DSTRING_MINBLOCKSZ || !path)
    return(NULL);
  if(!*path)
    return(NULL);
  if(stat(path, &st))
    return(NULL);
  if(!(fp = fopen(path, "r")))
    return(NULL);

  d = P_new(p_dstring_t);
  d->len = i = st.st_size;
  d->p = 0;
  d->blocksz = blocksz;
  d->blk = (d->len / d->blocksz) + 1;
  d->mem = p = P_newstr(d->blk * d->blocksz);
  while((b = fread(p, P_DSTRING_LOADBLOCKSZ, sizeof(char), fp))
        == P_DSTRING_LOADBLOCKSZ)
    i -= P_DSTRING_LOADBLOCKSZ, p += P_DSTRING_LOADBLOCKSZ;
  fclose(fp);

  if(b != i)
    {
    P_free(d->mem);
    P_free(d);
    return(NULL);
    }
  
  return(d);
  }

/*
 */

bool_t P_dstring_save(p_dstring_t *d, const char *path)
  {
  FILE *fp;
  uint_t i;
  char *p;
  size_t b;

  if(!d || !path)
    return(FALSE);
  if(!*path)
    return(FALSE);
  if(!(fp = fopen(path, "w")))
    return(FALSE);

  i = d->len, p = d->mem;
  while((b = fwrite(p, P_DSTRING_LOADBLOCKSZ, sizeof(char), fp))
        == P_DSTRING_LOADBLOCKSZ)
    i -= P_DSTRING_LOADBLOCKSZ, p += P_DSTRING_LOADBLOCKSZ;
  fclose(fp);
  
  return(b != i ? FALSE : TRUE);
  }

/* end of source file */
