/* ----------------------------------------------------------------------------
   Ping Utility Library
   Copyright (C) 1994-2000 PING Software Group and Mark A Lindner

   This file is part of pingutil.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   ----------------------------------------------------------------------------
   $Log: error.c,v $
   Revision 1.9  2000/12/31 03:57:11  markl
   Added new function.

   Revision 1.8  2000/08/17 07:23:30  markl
   Added new macro 'THREADED_LIBRARY' to correct some flaws in the
   configuration logic.

   Revision 1.7  2000/08/17 06:37:43  markl
   Changes to eliminate most compiler warnings.

   Revision 1.6  2000/08/10 07:59:11  markl
   Removed error serial count feature.

   Revision 1.5  2000/07/24 10:41:54  markl
   Fixed more thread-specific value logic problems.

   Revision 1.4  2000/07/22 11:38:11  markl
   Added new functions: strings, semaphores, and shared memory.

   Revision 1.3  2000/07/20 10:28:28  markl
   Fixed erroneous logic for retrieving thread-specific error value.

   Revision 1.2  2000/07/04 12:54:35  markl
   More fixes.

   Revision 1.1  2000/06/14 06:02:44  ping
   Initial checkin.

   ----------------------------------------------------------------------------
*/

/* Feature test switches */

#include "config.h"

/* System headers */

#ifdef THREADED_LIBRARY
#include <pthread.h>
#endif /* THREADED_LIBRARY */

/* Local headers */

#include "pingutil.h"
#include "perrno.h"

/* File scope variables */

static char *__P_error_progname = "<program>";

#ifdef THREADED_LIBRARY

static pthread_once_t __P_error_once = PTHREAD_ONCE_INIT;
static pthread_key_t __P_error_key;

#else

int p_errno = 0; /* global PING error number */

#endif /* THREADED_LIBRARY */

/* Functions */

void P_error_init(const char *progname)
  {
  if(progname)
    __P_error_progname = (char *)progname;
  }

/*
 */

void P_error_printf(const char *fmt, ...)
  {
  va_list vp;

  fprintf(stderr, "%s: ", __P_error_progname);
  va_start(vp, fmt);
  vfprintf(stderr, fmt, vp);
  va_end(vp);
  fflush(stderr);
  }

/*
 */

void P_error_print_usage(const char *usage)
  {
  
  fprintf(stderr, "%s: Usage: %s %s\n", __P_error_progname,
          __P_error_progname, usage);
  fflush(stderr);
  }

/*
 */

void P_error_syserr(void)
  {

  fprintf(stderr, "%s: %s\n", __P_error_progname, strerror(errno));
  fflush(stderr);
  }

/*
 */

#ifdef THREADED_LIBRARY

static void __P_error_destructor(void *datum)
  {
  P_free(datum);
  }

static void __P_error_init_once(void)
  {
  pthread_key_create(&__P_error_key, __P_error_destructor);
  }

/*
 */

int P_error_get_errno(void)
  {
  int *e;
  
  pthread_once(&__P_error_once, __P_error_init_once);

  e = (int *)pthread_getspecific(__P_error_key);
  if(!e)
    {
    /* didn't exist, so create it */

    e = P_new(int);
    *e = P_EOK;
    pthread_setspecific(__P_error_key, (void *)e);
    }

  return(*e);
  }

/*
 */

void P_error_set_errno(int err)
  {
  int *e;
  
  pthread_once(&__P_error_once, __P_error_init_once);

  e = (int *)pthread_getspecific(__P_error_key);
  if(!e)
    {
    /* didn't exist, so create it */

    e = P_new(int);    
    pthread_setspecific(__P_error_key, (void *)e);
    }

  *e = err;
  }

#endif /* THREADED_LIBRARY */

/* end of source file */
