/* ----------------------------------------------------------------------------
   Ping Utility Library
   Copyright (C) 1994-2000 PING Software Group and Mark A Lindner

   This file is part of pingutil.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   ----------------------------------------------------------------------------
   $Log: file.c,v $
   Revision 1.8  2000/08/17 07:23:30  markl
   Added new macro 'THREADED_LIBRARY' to correct some flaws in the
   configuration logic.

   Revision 1.7  2000/08/14 10:15:17  markl
   Fixed memory leak for linux.

   Revision 1.6  2000/08/13 05:56:42  markl
   More Linux ports.

   Revision 1.5  2000/08/10 07:59:29  markl
   Corrected readdir_r() call logic for non-POSIX version of the routine.

   Revision 1.4  2000/07/04 12:54:36  markl
   More fixes.

   Revision 1.3  2000/06/29 00:38:55  markl
   Minor fixes.

   Revision 1.2  2000/06/27 22:28:58  markl
   Bug fix related to POSIX version of readdir_r().

   Revision 1.1  2000/06/14 06:02:44  ping
   Initial checkin.

   ----------------------------------------------------------------------------
*/

/* Feature test switches */

#include "config.h"

/* System headers */

#include <sys/types.h>
#include <fcntl.h>
#include <dirent.h>

/* Local headers */

#include "pingutil.h"

/* Macros */

#define P_FILE_BLOCKSZ 15
#define P_FILE_PATHBLOCKSZ 64

/* Functions */

bool_t P_file_readdir(const char *path, struct p_dirlist *dir, int flags)
  {
  p_vector_t *vf, *vd;
  char *p;
  struct stat st;
  DIR *dp;
  struct dirent *de, *rde;
  int sz;

  if(!path || !dir)
    return(FALSE);
  if(chdir(path))
    return(FALSE);
  if(!(dp = opendir(".")))
    return(FALSE);

  sz = sizeof(struct dirent) + pathconf(".", _PC_NAME_MAX) + 1;
  
  vf = P_vector_start(P_FILE_BLOCKSZ);
  if(flags & P_FILE_SEP) vd = P_vector_start(P_FILE_BLOCKSZ);
  dir->nfiles = dir->ndirs = 0;

#if (defined(_POSIX_PTHREAD_SEMANTICS) || defined(linux) || defined(__hpux))
  de = (struct dirent *)P_malloc(sz, byte_t);

#ifdef OLD
 while(readdir_r(dp, &rde) == 0)
#else
 while(readdir_r(dp, de, &rde) == 0)
#endif

    {
    if(! rde)
      break;
#else
  rde = (struct dirent *)P_malloc(sz, byte_t);

  while((de = readdir_r(dp, rde)) != NULL)
    {
#endif
    
    if(lstat(de->d_name, &st)) continue;
    if(S_ISDIR(st.st_mode))
      {
      if(*(de->d_name) == '.')
	{
	if(!strcmp(de->d_name, "."))
	  {
          if(!(flags & P_FILE_SHOWDOT))
            continue;
          }
	else if(!strcmp(de->d_name, ".."))
	  {
          if(!(flags & P_FILE_SHOW2DOT))
            continue;
          }
	else if(!(flags & P_FILE_SHOWHIDDEN))
          continue;
	}
      p = ((flags & P_FILE_ADDSLASH)
           ? P_string_duptermin(de->d_name, '/') : P_string_dup(de->d_name));
      }
    else if(S_ISREG(st.st_mode))
      {
      if((*de->d_name == '.') && !(flags & P_FILE_SHOWHIDDEN))
        continue;
      p = P_string_dup(de->d_name);
      }
    else
      continue;  /* skip all other types of files */

    if(S_ISDIR(st.st_mode) && (flags & P_FILE_SEP))
      P_vector_store(vd, p);
    else
      P_vector_store(vf, p);
    }
  closedir(dp);

  dir->files = P_vector_end(vf, &(dir->nfiles));
  if(flags & P_FILE_SEP)
    dir->dirs = P_vector_end(vd, &(dir->ndirs));

  if(flags & P_FILE_SORT)
    {
    if(dir->nfiles)
      P_string_sortvec(dir->files, dir->nfiles);
    if(dir->ndirs)
      P_string_sortvec(dir->dirs, dir->ndirs);
    }

#if (defined(_POSIX_PTHREAD_SEMANTICS) || defined(linux))
  P_free(de);
#else
  P_free(rde);
#endif
  
  return(TRUE);
  }

/*
 */

static bool_t __P_file_descend(const char *path,
                               bool_t (*examine)(const char *file,
                                                 const struct stat *fst,
                                                 uint_t depth, void *hook),
                               uint_t depth, void *hook)
  {
  char *opath;
  struct stat fst;
  DIR *dp;
  struct dirent *de, *rde;
  bool_t r = TRUE;
  int sz;
 
  if(!(opath = P_file_getcwd())) return(FALSE);

  if(chdir(path))
    {
    P_free(opath);
    return(examine(path, NULL, depth, hook));
    }

  if(!(dp = opendir(".")))
    {
    chdir(opath);
    P_free(opath);
    return(examine(path, NULL, depth, hook));
    }

  sz = sizeof(struct dirent) + pathconf(".", _PC_NAME_MAX) + 1;



  
#if (defined(_POSIX_PTHREAD_SEMANTICS) || defined(linux) || defined(__hpux))
  de = (struct dirent *)P_malloc(sz, byte_t);

#ifdef OLD
 while(readdir_r(dp, &rde) == 0)
#else
 while(readdir_r(dp, de, &rde) == 0)
#endif

    {
    if(! rde)
      break;
#else
    rde = (struct dirent *)P_malloc(sz, byte_t);

    while((de = readdir_r(dp, rde)) != NULL)
      {
#endif
    
    if(!strcmp(de->d_name, ".") || !strcmp(de->d_name, ".."))
      continue;
    lstat(de->d_name, &fst);
    if(S_ISDIR(fst.st_mode))
      {
      if(!examine(de->d_name, &fst, depth, hook))
	{
	r = FALSE;
	break;
	}
      
      if(!__P_file_descend(de->d_name, examine, depth + 1, hook))
	{
	r = FALSE;
	break;
	}
      }
    else if(S_ISREG(fst.st_mode))
      {
      if(!examine(de->d_name, &fst, depth, hook))
	{
	r = FALSE;
	break;
	}
      }
    else continue; /* skip all other types of files */
    }
  closedir(dp);
  chdir(opath);
  P_free(opath);
#if (defined(_POSIX_PTHREAD_SEMANTICS) || defined(linux))
  P_free(de);
#else
  P_free(rde);
#endif
  
  return(r);
  }

/*
 */

bool_t P_file_traverse(const char *path,
                       bool_t (*examine)(const char *file,
                                         const struct stat *fst,
                                         uint_t depth, void *hook),
                       void *hook)
  {
  if(!path || !examine)
    return(FALSE);
  if(!*path)
    return(FALSE);

  return(__P_file_descend(path, examine, 0, hook));
  }

/*
 */

char *P_file_getcwd(void)
  {
  char *p = P_newstr(P_FILE_PATHBLOCKSZ);
  size_t bufsz = P_FILE_PATHBLOCKSZ, l;

  for(;;)
    {
    if(getcwd(p, bufsz - 1))
      {
      l = strlen(p);
      *(p + l) = NUL;
      return(P_realloc(p, l + 1, char));
      }
    if(errno != ERANGE)
      {
      P_free(p);
      return(NULL);
      }
    bufsz += P_FILE_PATHBLOCKSZ;
    p = P_realloc(p, bufsz, char);
    continue;
    }
  }

/*
 */

bool_t P_file_lock(FILE *fp, int type)
  {
  int r;
  int t = (type == P_FILE_READ_LOCK) ? F_RDLCK : F_WRLCK;
#if defined (__hpux) || defined (linux)
  struct flock lock = { t, SEEK_SET, 0, 0, 0 };
#else
  struct flock lock = { t, SEEK_SET, 0, 0, 0, 0 };
#endif  
  int fd = fileno(fp);

  r = fcntl(fd, F_SETLKW, lock);

  return(r >= 0);
  }

/*
 */

bool_t P_file_trylock(FILE *fp, int type)
  {
  int r;
  int t = (type == P_FILE_READ_LOCK) ? F_RDLCK : F_WRLCK;
#if defined (__hpux) || defined (linux)
  struct flock lock = { t, SEEK_SET, 0, 0, 0 };
#else
  struct flock lock = { t, SEEK_SET, 0, 0, 0, 0 };
#endif
  int fd = fileno(fp);

  r = fcntl(fd, F_SETLKW, lock);

  return(r >= 0);
  }

/*
 */

bool_t P_file_unlock(FILE *fp, int type)
  {
  int r;
  int t = (type == P_FILE_READ_LOCK) ? F_RDLCK : F_WRLCK;
#if defined (__hpux) || defined (linux)
  struct flock lock = { t, SEEK_SET, 0, 0, 0 };
#else
  struct flock lock = { t, SEEK_SET, 0, 0, 0, 0 };
#endif
  int fd = fileno(fp);

  r = fcntl(fd, F_SETLK, lock);

  return(r >= 0);
  }

/* end of source file */
