/* ----------------------------------------------------------------------------
   Ping Utility Library
   Copyright (C) 1994-2000 PING Software Group and Mark A Lindner

   This file is part of pingutil.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   ----------------------------------------------------------------------------
   $Log: hashtab.c,v $
   Revision 1.4  2000/12/13 11:28:35  markl
   Bug fixes.

   Revision 1.3  2000/12/08 20:03:00  markl
   Bug fixes, and added new hashtable function.

   Revision 1.2  2000/07/04 12:54:36  markl
   More fixes.

   Revision 1.1  2000/06/14 06:02:44  ping
   Initial checkin.

   ----------------------------------------------------------------------------
*/

/* Feature test switches */

#include "config.h"

/* Local headers */

#include "pingutil.h"

/* File scope variables */

static uint_t (*__P_hashtable_hashfunc)(const char *s, uint_t modulo)
  = P_string_hash;

/* Functions */

bool_t P_hashtable_set_hashfunc(uint_t (*func)(const char *s, uint_t modulo))
  {
  bool_t r = FALSE;
  
  if(func)
    __P_hashtable_hashfunc = func, r = TRUE;

  return(r);
  }

/*
 */

p_hashtable_t *P_hashtable_create(uint_t buckets)
  {
  p_hashtable_t *h;

  if(!buckets)
    return(NULL);

  h = P_new(p_hashtable_t);
  h->buckets = buckets;
  h->size = 0;
  h->table = P_calloc(buckets, p_linklist_t *);

  return(h);
  }

/*
 */

void P_hashtable_destroy(p_hashtable_t *h)
  {
  uint_t i;
  p_linklist_t **p;

  if(!h)
    return;

  /* this dereferencing may be buggy */

  for(i = h->buckets, p = h->table; i--; p++)
    P_linklist_destroy(*p); /* if NULL, harmless */
  P_free(h->table);
  P_free(h);
  }

/*
 */

bool_t P_hashtable_store(p_hashtable_t *h, const char *key, const void *data)
  {
  p_tag_t *tag;
  p_linklist_t **l;

  if(!h || !key || !data)
    return(FALSE);
  if(!*key)
    return(FALSE);

  tag = P_new(p_tag_t);
  tag->key = (char *)key;
  tag->data = (char *)data;

  l = &(h->table[__P_hashtable_hashfunc(key, h->buckets)]);
  
  if(!(*l))
    *l = P_linklist_create();

  P_linklist_push(*l, (void *)tag);
  h->size++;
  
  return(TRUE);
  }

/*
 */

void *P_hashtable_restore(p_hashtable_t *h, const char *key)
  {
  p_tag_t *tag;
  p_linklist_t *l;

  if(!h || !key) return(FALSE);
  if(!*key) return(FALSE);

  if((l = h->table[__P_hashtable_hashfunc(key, h->buckets)]))
    {
    for(P_linklist_move_head(l); !P_linklist_istail(l);
        P_linklist_move_next(l))
      {
      tag = (p_tag_t *)P_linklist_restore(l);
      if(!strcmp(tag->key, key))
        return(tag->data);
      }
    }

  return(NULL);
  }

/*
 */

bool_t P_hashtable_delete(p_hashtable_t *h, const char *key)
  {
  p_tag_t *tag;
  p_linklist_t *l;

  if(!*key) return(FALSE);

  if((l = h->table[__P_hashtable_hashfunc(key, h->buckets)]))
    {
    for(P_linklist_move_head(l); !P_linklist_istail(l);
        P_linklist_move_next(l))
      {
      tag = (p_tag_t *)P_linklist_restore(l);
      if(!strcmp(tag->key, key))
	{
	P_linklist_delete(l);
	h->size--;
	return(TRUE);
	}
      }
    }
  
  return(FALSE);
  }

/*
 */

char **P_hashtable_keys(p_hashtable_t *h, uint_t *len)
  {
  p_vector_t *vec;
  p_linklist_t **p;
  p_tag_t *t;
  char **v;
  uint_t i;

  if(!h)
    return(NULL);

  vec = P_vector_start(h->buckets);

  for(i = h->buckets, p = h->table; i--; p++)
    {
    for(P_linklist_move_head(*p);
        (t = (p_tag_t *)P_linklist_restore(*p)) != NULL;
        P_linklist_move_next(*p))
      {
      P_vector_store(vec, P_string_dup(t->key));
      }
    }

  v = P_vector_end(vec, len);

  return(v);
  }

/* end of source file */
