/* ----------------------------------------------------------------------------
   Ping Utility Library
   Copyright (C) 1994-2000 PING Software Group and Mark A Lindner

   This file is part of pingutil.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   ----------------------------------------------------------------------------
   $Log: pingnet.h,v $
   Revision 1.6  2001/01/21 08:50:44  markl
   Checks and conditional typedefs for in_addr_t and in_port_t

   Revision 1.5  2000/08/13 05:56:43  markl
   More Linux ports.

   Revision 1.4  2000/08/13 05:34:34  markl
   First cut at Linux port.

   Revision 1.3  2000/08/10 08:22:23  markl
   Added multicast functions.

   Revision 1.2  2000/06/29 00:38:55  markl
   Minor fixes.

   Revision 1.1  2000/06/14 06:02:46  ping
   Initial checkin.

   ----------------------------------------------------------------------------
*/

#ifndef __pingnet_h
#define __pingnet_h

#include "pingutil.h"

#include <sys/socket.h>
#include <netinet/in.h>

/* terminal control */

extern bool_t P_tty_raw(int fd);
extern bool_t P_tty_unraw(int fd);
extern bool_t P_tty_store(int fd);
extern bool_t P_tty_restore(int fd);
extern bool_t P_tty_generic(int fd);

/* pseudoterminals */

extern int P_pty_master(void);
extern int P_pty_slave(int mfd);
extern int P_pty_exec(const char *path, char *const argv[], pid_t *p);

/* sockets */

typedef struct p_socket_t
  {
  int sd;
  FILE *sfp;
  struct sockaddr_in laddr;
  struct sockaddr_in raddr;
  unsigned short flags;
  unsigned short state;
  unsigned short type;
  } p_socket_t;

/* socket control functions */

extern bool_t P_socket_create(p_socket_t *s, int type);
extern bool_t P_socket_destroy(p_socket_t *s);
extern bool_t P_socket_listen(p_socket_t *s, in_port_t port);
extern bool_t P_socket_accept(p_socket_t *s, p_socket_t *ns);
extern bool_t P_socket_connect(p_socket_t *s, const char *host,
                               in_port_t port);
extern bool_t P_socket_shutdown(p_socket_t *s, uint_t how);

extern bool_t P_socket_fopen(p_socket_t *s, int buffering);
extern bool_t P_socket_fclose(p_socket_t *s);
extern bool_t P_socket_reopen(int sd, p_socket_t *s);

extern bool_t P_socket_get_peeraddr(p_socket_t *s, char *buf, size_t bufsz);

extern bool_t P_socket_set_option(p_socket_t *s, uint_t option, bool_t flag,
                                  uint_t value);
extern bool_t P_socket_get_option(p_socket_t *s, uint_t option, bool_t *flag,
                                  uint_t *value);

/* buffering modes */

#define P_NET_BUFFERING_NONE _IONBF
#define P_NET_BUFFERING_LINE _IOLBF
#define P_NET_BUFFERING_FULL _IOFBF

/* socket shutdown states */

#define P_NET_SHUTRD 1
#define P_NET_SHUTWR 2
#define P_NET_SHUTALL 3

/* socket options */

#define P_NET_OPT_BLOCK 1
#define P_NET_OPT_LINGER 2
#define P_NET_OPT_REUSEADDR 3
#define P_NET_OPT_OOBINLINE 4
#define P_NET_OPT_KEEPALIVE 5
#define P_NET_OPT_RECVBUF 6
#define P_NET_OPT_SENDBUF 7

/* socket types */

#define P_NET_MAXTYPE 2
#define P_NET_NTYPES 3
#define P_NET_TCP 0
#define P_NET_UDP 1
#define P_NET_UNKNOWN 2
#define P_NET_OTHER P_NET_UNKNOWN

/* constants */

#define P_NET_BACKLOG 5

/* flags and masks */

#define P_NET_MSHUT 0x01
#define P_NET_MUNBLOCK 0x02
#define P_NET_OSHUT 0
#define P_NET_OSHUTRD 0
#define P_NET_OSHUTWR 1
#define P_NET_OUNBLOCK 2

/* socket control macros */

#define P_socket_set_iobufsz(S, I, O) \
  P_socket_iobufsz((S), (I), (O), TRUE)
#define P_socket_get_iobufsz(S, I, O) \
  P_socket_iobufsz((S), (I), (O), FALSE)
#define P_socket_block(S) \
  P_socket_set_option((S), P_NET_OPT_BLOCK, TRUE, 0)
#define P_socket_unblock(S) \
  P_socket_set_option((S), P_NET_OPT_BLOCK, FALSE, 0)
#define P_socket_get_fd(S) ((S)->sd)
#define P_socket_get_fp(S) ((S)->sfp)
#define P_socket_isblocked(S) \
  (((S)->flags & P_NET_MUNBLOCK) ? FALSE : TRUE)
#define P_socket_get_type(S) ((S)->type)
#define P_socket_get_ipaddr(S) \
	(in_addr_t)(nothl((S)->laddr.sin_addr.s_addr))
#define P_socket_get_peeripaddr(S) \
	(in_addr_t)(ntohl((S)->raddr.sin_addr.s_addr))

/* socket multicast functions */

extern bool_t P_socket_mcast_join(p_socket_t *s, const char *addr);
extern bool_t P_socket_mcast_leave(p_socket_t *s, const char *addr);

extern bool_t P_socket_mcast_set_ttl(p_socket_t *s, byte_t ttl);

/* socket I/O functions */

extern int P_socket_send(p_socket_t *s, const char *buf, size_t bufsz,
			 bool_t oobf);
extern int P_socket_recv(p_socket_t *s, char *buf, size_t bufsz, bool_t boof);

extern int P_socket_sendto(p_socket_t *s, const char *buf, size_t bufsz,
			   const char *addr, in_port_t port);
extern int P_socket_recvfrom(p_socket_t *s, char *buf, size_t bufsz,
			     char *addr, size_t addrsz);

extern int P_socket_sendreply(p_socket_t *s, const char *buf, size_t bufsz);
extern int P_socket_recvreply(p_socket_t *s, char *buf, size_t bufsz);

extern int P_socket_writeline(p_socket_t *s, const char *buf,
                              const char *termin, uint_t slen, uint_t snum);
extern int P_socket_readline(p_socket_t *s, char *buf, size_t bufsz,
                             char termin, uint_t slen, uint_t snum);

#define P_NET_DFL_SLEN 30
#define P_NET_DFL_SNUM 10

/* socket I/O macros */

#define P_socket_wl(S, B, T) \
  P_socket_writeline((S), (B), (T), P_NET_DFL_SLEN, P_NET_DFL_SNUM)
#define P_socket_rl(S, B, BS, T) \
  P_socket_readline((S), (B), (BS), (T), P_NET_DFL_SLEN, P_NET_DFL_SNUM)

/* network information functions */

extern in_port_t P_net_get_svcport(const char *name, uint_t *type);
extern bool_t P_net_get_svcname(in_port_t port, uint_t *type, char *buf,
				size_t bufsz);
extern bool_t P_net_resolve(const char *ipaddr, char *buf, size_t bufsz);
extern bool_t P_net_resolve_local(char *addr, char *ipaddr, size_t bufsz,
				  in_addr_t *ip);

#endif /* __pingnet_h */

/* end of library header */

