/* ----------------------------------------------------------------------------
   Ping Utility Library
   Copyright (C) 1994-2000 PING Software Group and Mark A Lindner

   This file is part of pingutil.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   ----------------------------------------------------------------------------
   $Log: sched.c,v $
   Revision 1.8  2000/08/17 07:23:30  markl
   Added new macro 'THREADED_LIBRARY' to correct some flaws in the
   configuration logic.

   Revision 1.7  2000/08/17 01:01:00  markl
   Range checking bug fix.

   Revision 1.6  2000/08/13 06:02:01  markl
   #ifdef'ed out all of the scheduler functions for Linux, since there is no
   POSIX.4 timer support on Linux.

   Revision 1.5  2000/07/04 12:54:38  markl
   More fixes.

   Revision 1.4  2000/07/04 12:16:55  markl
   Autoconf enabled.

   Revision 1.3  2000/06/26 20:21:24  markl
   *** empty log message ***

   Revision 1.2  2000/06/15 22:47:00  markl
   Added IDs to p_schedevt_t, added find function.

   Revision 1.1  2000/06/14 06:02:47  ping
   Initial checkin.

   ----------------------------------------------------------------------------
*/

/* Feature test switches */

#ifndef linux /* no POSIX real-time timer support on Linux */

#include "config.h"

#define _POSIX_C_SOURCE 199506L

/* System headers */

#ifdef THREADED_LIBRARY
#include <pthread.h>
#endif /* THREADED_LIBRARY */

/* Local headers */

#include "pingutil.h"

/* Macros */

#define ONE_MINUTE 60 /* seconds in a minute */

#define P_SCHED_INTERVAL ONE_MINUTE

#define P_SCHED_MINUTE_MASK 0
#define P_SCHED_HOUR_MASK 1
#define P_SCHED_DATE_MASK 2
#define P_SCHED_MONTH_MASK 3
#define P_SCHED_WEEKDAY_MASK 4

/* File scope variables */

static struct sigaction __P_sched_old_sigaction;
static void (*__P_sched_event_handler)(p_schedevt_t *);
static p_sched_t *__P_sched_scheduler;

static const int __P_sched_range_max[] = { 59, 23, 31, 12, 6 };
static const int __P_sched_range_min[] = { 0, 0, 1, 1, 0 };

#ifdef THREADED_LIBRARY
static pthread_t __P_sched_wait_thread;
#endif /* THREADED_LIBRARY */

/* File scope functions */

static bool_t __P_sched_event_deactivate(p_sched_t *s, p_schedevt_t *e)
  {
  if(!s || !e)
    return(FALSE);

  if(! P_linklist_search(s->events, (void *)e))
    return(FALSE); /* not found */

  if(! P_linklist_delete(s->events))
    return(FALSE); /* delete failed (?) */

  if(e->destructor)
    e->destructor(e);
  
  return(TRUE);
  }

/*
 */

static bool_t __P_sched_event_activate(p_sched_t *s, p_schedevt_t *e)
  {
  if(!s || !e)
    return(FALSE);

  if(P_linklist_search(s->events, (void *)e))
    return(FALSE); /* found it already in there*/

  P_linklist_move_tail(s->events);
  
  if(! P_linklist_store(s->events, (void *)e))
    return(FALSE); /* store failed (?)*/

  return(TRUE);
  }

/*
 */

static void __P_sched_default_handler(p_schedevt_t *evt)
  {
  time_t t = time(NULL);
  struct tm now;
  uint_t tval[P_SCHED_MASK_COUNT];
  
  localtime_r(&t, &now);
  
  tval[P_SCHED_MINUTE_MASK] = (uint_t)now.tm_min;
  tval[P_SCHED_HOUR_MASK] = (uint_t)now.tm_hour;
  tval[P_SCHED_DATE_MASK] = (uint_t)now.tm_mday;
  tval[P_SCHED_MONTH_MASK] = (uint_t)now.tm_mon;
  tval[P_SCHED_WEEKDAY_MASK] = (uint_t)now.tm_wday;

#ifdef DEBUG
  printf("mn = %d, hr = %d, dt = %d, mo = %d, wd = %d\n", tval[0],
         tval[1], tval[2], tval[3], tval[4]);
#endif

  for(P_linklist_move_head(__P_sched_scheduler->events);
      ! P_linklist_istail(__P_sched_scheduler->events);
      P_linklist_move_next(__P_sched_scheduler->events))
    {
    int i;
    bool_t trigger = TRUE;
    p_schedevt_t *evt
      = (p_schedevt_t *)P_linklist_restore(__P_sched_scheduler->events);
    
    for(i = 0; i < P_SCHED_MASK_COUNT; i++)
      {
      if(! P_bitstring_isset(evt->mask[i], tval[i]))
        {
        trigger = FALSE;
        break;
        }
      }

    if(trigger)
      {
#ifdef DEBUG
      printf(ctime(&t));
#endif

      if(evt->handler)
        evt->handler(evt, t);

      if(evt->once)
        {
        __P_sched_event_deactivate(__P_sched_scheduler, evt);
        }
      }
    }
  }

/*
 */

#ifdef THREADED_LIBRARY

static void *__P_sched_threaded_handler(void *arg)
  {
  sigset_t sigmask;
  int sig, old;

  sigemptyset(&sigmask);
  sigaddset(&sigmask, SIGRTMIN);

  pthread_sigmask(SIG_UNBLOCK, &sigmask, NULL);

  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  
  for(;;)
    {
    sigwait(&sigmask, &sig);

    pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &old);
    
    if(sig == SIGRTMIN)
      __P_sched_default_handler(NULL);

    pthread_setcancelstate(old, NULL);
    }
  }

#endif /* THREADED_LIBRARY */

/*
 */

static void __P_sched_internal_handler(int sig, siginfo_t *extra, void *cruft)
  {
#ifndef THREADED_LIBRARY
  p_schedevt_t *evt = NULL;
#endif
    
  /*
    int overrun = 0;

    evt = (schedevt_t *)(extra->si_value.sival_ptr);
    overrun = timer_getoverrun(sched->timer);

    printf("si_signo = %d\n", extra->si_signo);
    printf("si_errno = %d\n",  extra->si_errno);
    printf("si_code  = %d\n", extra->si_code);
    printf("si_value.sival_int = %d\n", extra->si_value.sival_int);
  
    printf("data value is %d\n", extra->si_value.sival_int);
  */

#ifndef THREADED_LIBRARY
  
  if(__P_sched_event_handler)
    __P_sched_event_handler(evt);

#endif /* ! THREADED_LIBRARY */
  }

/*
 */

static bool_t __P_sched_parsetime(char *timespec, p_bitstring_t *mask,
                                  uint_t min)
  {
  char *p, *q, *pp, *qq, buf[256];
  int start, end, x = 0;

  if(!strcmp(timespec, "*"))
    P_bitstring_set_all(mask);

  else
    {
    P_bitstring_clear_all(mask);

    q = NULL;
    
    for(p = strtok_r(timespec, ",", &q);
        p;
        p = strtok_r(NULL, ",", &q), x++)
      {
      start = end = -1;

      strncpy(buf, p, sizeof(buf) - 1);
      buf[sizeof(buf) - 1] = NUL;
      qq = NULL;

      if(!((pp = strtok_r(buf, "-", &qq))))
        continue;

      if(! P_string_isnumeric(pp))
        return(FALSE); /* parse error! 8/

      start = atoi(pp);

      if((pp = strtok_r(NULL,  "-", &qq)))
        {
        if(! P_string_isnumeric(pp))
          return(FALSE); /* parse error! */

        end = atoi(pp);
        }

      if((start < min) || (start >= P_bitstring_size(mask)))
        return(FALSE); /* out of range! */
      
      if(end < 0)
        P_bitstring_set(mask, (uint_t)start);

      else if(end < start)
        return(FALSE); /* invalid range! */

      else if((end < min) || (end >= P_bitstring_size(mask)))
        return(FALSE); /* out of range! */

      else
        P_bitstring_set_range(mask, (uint_t)start, (uint_t)end);
      }

    if(x == 0)
     { return(FALSE); /* empty input */}
    else
     { return(TRUE); }
  }

/*
 */

static p_sched_t *__P_sched_create(void *hook)
  {
  time_t t = time(NULL);
  struct itimerspec interval;
  p_sched_t *sched;
  struct tm now;

  localtime_r(&t, &now);
  
  sched = P_new(p_sched_t);
  
  sched->hook = hook;
  sched->evp.sigev_notify = SIGEV_SIGNAL;
  sched->evp.sigev_signo = SIGRTMIN;
  sched->evp.sigev_value.sival_int = 0; /* buggy in Solaris 7 */

  sched->events = P_linklist_create();

  timer_create(CLOCK_REALTIME, &(sched->evp), &(sched->timer));

  interval.it_value.tv_sec = ONE_MINUTE - now.tm_sec;
  interval.it_value.tv_nsec = 0;

  interval.it_interval.tv_sec = P_SCHED_INTERVAL;
  interval.it_interval.tv_nsec = 0;

  timer_settime(sched->timer, 0, &interval, NULL);

  return(sched);
  }
 
/*
 */

static bool_t __P_sched_destroy(p_sched_t *s)
  {
  if(!s)
    return(FALSE);
  
  timer_delete(s->timer);

  /* clean up the linked list of events */

  P_linklist_destroy(s->events);

  P_free(s);

  return(TRUE);
  }

/*
 */

 static void __P_sched_set_handler(void (*func)(p_schedevt_t *))
  {
  __P_sched_event_handler = func;
  }

/* Functions */

bool_t P_sched_init(void)
  {
  struct sigaction sa;

  if(__P_sched_scheduler)
    return(FALSE); /* scheduler is already initialized! */
  
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = SA_SIGINFO;
  sa.sa_sigaction = __P_sched_internal_handler;

  if(sigaction(SIGRTMIN, &sa, &__P_sched_old_sigaction) < 0)
    return(FALSE);

  __P_sched_set_handler(__P_sched_default_handler);

  __P_sched_scheduler = __P_sched_create(NULL);

#ifdef THREADED_LIBRARY
  pthread_create(&__P_sched_wait_thread, NULL, __P_sched_threaded_handler,
                 NULL);
#endif /* THREADED_LIBRARY */
  
  return(TRUE);
  }

/*
 */

p_schedevt_t *P_sched_event_find(uint_t id)
  {
  for(P_linklist_move_head(__P_sched_scheduler->events);
      ! P_linklist_istail(__P_sched_scheduler->events);
      P_linklist_move_next(__P_sched_scheduler->events))
    {
    p_schedevt_t *evt
      = (p_schedevt_t *)P_linklist_restore(__P_sched_scheduler->events);

    if(evt->id == id)
      return(evt);
    }

  return(NULL);
  }

/*
 */

bool_t P_sched_shutdown(void)
  {
  if(! __P_sched_scheduler)
    return(FALSE);
  
  sigaction(SIGRTMIN, &__P_sched_old_sigaction, NULL);

#ifdef THREADED_LIBRARY
  pthread_cancel(__P_sched_wait_thread);
#endif /* THREADED_LIBRARY */
  
  for(P_linklist_move_head(__P_sched_scheduler->events);
      ! P_linklist_istail(__P_sched_scheduler->events);
      P_linklist_move_next(__P_sched_scheduler->events))
    {
    p_schedevt_t *evt
      = (p_schedevt_t *)P_linklist_restore(__P_sched_scheduler->events);

    __P_sched_event_deactivate(__P_sched_scheduler, evt);
    }

  __P_sched_destroy(__P_sched_scheduler);

  __P_sched_scheduler = NULL;

  return(TRUE);
  }

/*
 */

p_schedevt_t *P_sched_event_create(const char *timespec, bool_t once,
                                   void *hook,
                                   void (*handler)(p_schedevt_t *, time_t),
                                   void (*destructor)(p_schedevt_t *),
                                   uint_t id)
  {
  p_schedevt_t *evt;
  char *p, *q = NULL, buf[256];
  int i;

  evt = P_new(p_schedevt_t);
  evt->once = once;
  evt->hook = hook;
  evt->handler = handler;
  evt->destructor = destructor;
  evt->id = id;

  strncpy(buf, timespec, sizeof(buf) - 1);
  buf[sizeof(buf) - 1] = NUL;

  for(i = 0, p = strtok_r(buf, " :\t", &q);
      (i < P_SCHED_MASK_COUNT) && p;
      i++, p = strtok_r(NULL, " :\t", &q))
    {
    evt->mask[i] = P_bitstring_create(__P_sched_range_max[i] + 1);
    if(!(__P_sched_parsetime(p, evt->mask[i], __P_sched_range_min[i])))
      break;
    }

  if(i != P_SCHED_MASK_COUNT)
    {
    P_sched_event_destroy(evt);
    return(NULL);
    }

  return(evt);
  }

/*
 */

bool_t P_sched_event_destroy(p_schedevt_t *evt)
  {
  int i;
  
  if(!evt)
    return(FALSE);

  for(i = 0; i < P_SCHED_MASK_COUNT; i++)
    P_free(evt->mask[i]);
  
  P_free(evt);
  
  return(TRUE);
  }

/*
 */

bool_t P_sched_event_activate(p_schedevt_t *e)
  {
  if(! __P_sched_scheduler)
    return(FALSE);

  return(__P_sched_event_activate(__P_sched_scheduler, e));
  }

/*
 */

bool_t P_sched_event_deactivate(p_schedevt_t *e)
  {
  if(! __P_sched_scheduler)
    return(FALSE);
  
  return(__P_sched_event_deactivate(__P_sched_scheduler, e));
  }

#endif /* ! linux */

/* end of source file */

