/* ----------------------------------------------------------------------------
   Ping Utility Library
   Copyright (C) 1994-2000 PING Software Group and Mark A Lindner

   This file is part of pingutil.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   ----------------------------------------------------------------------------
   $Log: sockctl.c,v $
   Revision 1.7  2000/12/08 20:03:01  markl
   Bug fixes, and added new hashtable function.

   Revision 1.6  2000/09/23 09:28:15  markl
   Removed erroneous htonl() call.

   Revision 1.5  2000/08/17 06:37:45  markl
   Changes to eliminate most compiler warnings.

   Revision 1.4  2000/08/13 05:34:34  markl
   First cut at Linux port.

   Revision 1.3  2000/08/10 08:22:23  markl
   Added multicast functions.

   Revision 1.2  2000/07/04 12:54:38  markl
   More fixes.

   Revision 1.1  2000/06/14 06:02:48  ping
   Initial checkin.

   ----------------------------------------------------------------------------
*/

/* Feature test switches */

#include "config.h"

/* System headers */

#include <fcntl.h>
#include <netinet/in.h>
#include <ctype.h>


#ifdef __hpux
/*
 *  From in.h  ! _XOPEN_SOURCE_EXTENDED:
 * Argument structure for IP_ADD_MEMBERSHIP and IP_DROP_MEMBERSHIP.
 */
struct ip_mreq {
        struct in_addr  imr_multiaddr;  /* IP multicast address of group */
        struct in_addr  imr_interface;  /* local IP address of interface */
};
#endif

/* Local headers */

#include "pingutil.h"
#include "pingnet.h"
#include "perrno.h"
#include "netcommon.h"

/* File scope variables */

const int __P_net_socktypes[P_NET_NTYPES] =
  { SOCK_STREAM, SOCK_DGRAM, -1 };

const char *__P_net_protocols[P_NET_NTYPES] =
  { "tcp", "udp", NULL };

/* External functions */

bool_t __P_socket_addr2sock(struct sockaddr_in *sa, const char *addr)
  {
  struct hostent he;
#ifdef linux
  struct hostent *rhe;
#endif
  p_buffer_t *rbuf = __P_net_get_buffer();
  int herr;

  if(!sa || !addr)
    return(FALSE);

  if(isdigit((int)*addr))
    {
    if((sa->sin_addr.s_addr = inet_addr(addr)) == -1)
      return(FALSE);
    sa->sin_family = AF_INET;
    }
  else
    {
GETHOSTBYNAME:
#ifdef linux
    if(gethostbyname_r(addr, &he, rbuf->buf, rbuf->bufsz, &rhe, &herr))
#else
    if(! gethostbyname_r(addr, &he, rbuf->buf, rbuf->bufsz, &herr))
#endif
      {
      if(herr == ERANGE)
        {
        P_buffer_resize(rbuf, rbuf->bufsz + P_NET_BUFSZ);
        goto GETHOSTBYNAME;
        }
      else
        {
        P_error_set_errno(P_EADDRINFO);
        return(FALSE);
        }
      }

    memcpy((void *)&(sa->sin_addr), (void *)he.h_addr, (size_t)he.h_length);
    sa->sin_family = he.h_addrtype;
    }
  
  return(TRUE);
  }

/*
 */

bool_t __P_socket_sock2addr(struct sockaddr_in *sa, char *addr,
			   size_t addrsz)
  {
  struct hostent he;
#ifdef linux
  struct hostent *rhe;
#endif
  p_buffer_t *rbuf = __P_net_get_buffer();
  int herr;

  if(!sa || !addr || !addrsz)
    return(FALSE);

GETHOSTBYADDR:
#ifdef linux
  if(! gethostbyaddr_r((char *)&(sa->sin_addr.s_addr), sizeof(in_addr_t),
                       AF_INET, &he, rbuf->buf, rbuf->bufsz, &rhe, &herr))
#else
  if(! gethostbyaddr_r((char *)&(sa->sin_addr.s_addr), sizeof(in_addr_t),
                       AF_INET, &he, rbuf->buf, rbuf->bufsz, &herr))
#endif
    {
    if(herr == ERANGE)
      {
      P_buffer_resize(rbuf, rbuf->bufsz + P_NET_BUFSZ);
      goto GETHOSTBYADDR;
      }
    else
      {    
      P_error_set_errno(P_EADDRINFO);
      return(FALSE);
      }
    }

  strncpy(addr, he.h_name, --addrsz);
  *(addr + addrsz) = NUL;

  return(TRUE);
  }

/*
 */

bool_t P_socket_create(p_socket_t *s, int type)
  {
  if((type < 0) || (type >= P_NET_MAXTYPE) || !s)
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }

  s->sfp = NULL;
  s->flags = 0;
  s->state = P_NET_CREATED;
  s->type = type;

  if((s->sd = socket(AF_INET, __P_net_socktypes[type], 0)) < 0)
    {
    P_error_set_errno(P_ESOCKET);
    return(FALSE);
    }
  
  return(TRUE);
  }

/*
 */

bool_t P_socket_listen(p_socket_t *s, in_port_t port)
  {
  struct hostent he;
#ifdef linux
  struct hostent *rhe;
#endif
  p_buffer_t *rbuf = __P_net_get_buffer();
  int herr, x = 1;
  char *myname;

  if(!s)
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }
  
  if(s->state != P_NET_CREATED)
    {
    P_error_set_errno(P_EBADSTATE);
    return(FALSE);
    }
  
  myname = P_system_get_hostname();

GETHOSTBYNAME2:
#ifdef linux
  if(gethostbyname_r(myname, &he, rbuf->buf, rbuf->bufsz, &rhe, &herr))
#else
  if(! gethostbyname_r(myname, &he, rbuf->buf, rbuf->bufsz, &herr))
#endif
    {
    if(herr == ERANGE)
      {
      P_buffer_resize(rbuf, rbuf->bufsz + P_NET_BUFSZ);
      goto GETHOSTBYNAME2;
      }
    else
      {
      P_error_set_errno(P_EADDRINFO);
      return(FALSE);
      }
    }

  s->laddr.sin_addr.s_addr = htonl(INADDR_ANY);
  s->laddr.sin_family = he.h_addrtype;
  s->laddr.sin_port = htons(port);

  if(setsockopt(s->sd, SOL_SOCKET, SO_REUSEADDR, (void *)&x, sizeof(int))
     < 0)
    {
    P_error_set_errno(P_ESOCKINFO);
    return(FALSE);
    }
  
  if(bind(s->sd, (struct sockaddr *)&(s->laddr), sizeof(struct sockaddr_in))
     < 0)
    {
    P_error_set_errno(P_EBIND);
    return(FALSE);
    }

  if(s->type == P_NET_TCP)
    {    
    if(listen(s->sd, P_NET_BACKLOG) < 0)
      {
      P_error_set_errno(P_ELISTEN);
      return(FALSE);
      }
    s->state = P_NET_LISTENING;
    }

  return(TRUE);
  }

/*
 */

bool_t P_socket_accept(p_socket_t *s, p_socket_t *ns)
  {
  int sz = sizeof(struct sockaddr_in);

  if(!s || !ns)
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }
  
  if(s->state != P_NET_LISTENING)
    {
    P_error_set_errno(P_EBADSTATE);
    return(FALSE);
    }

  if(s->type != P_NET_TCP)
    {
    P_error_set_errno(P_EBADTYPE);
    return(FALSE);
    }

ACCEPT:
  if((ns->sd = accept(s->sd, (struct sockaddr *)&(ns->raddr), &sz)) < 0)
    {
    switch(errno)
      {
      case EWOULDBLOCK:
        P_error_set_errno(P_EBLOCKED);
	return(FALSE);
        
      case EINTR:
      case ECONNABORTED:
	goto ACCEPT;
        
      default:
	P_error_set_errno(P_EACCEPT);
	return(FALSE);
      }
    }
  
  ns->type = s->type;
  ns->state = P_NET_CONNECTED;
  ns->flags = s->flags;
  ns->sfp = NULL;

  ns->flags &= ~(P_NET_MUNBLOCK);

  memcpy((void *)&(ns->laddr), (void *)&(s->laddr),
         sizeof(struct sockaddr_in));
  
  return(TRUE);
  }

/*
 */

bool_t P_socket_connect(p_socket_t *s, const char *host, in_port_t port)
  {
  int sz = (int)sizeof(struct sockaddr_in);

  if(!s || !host)
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }

  if(! *host)
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }
  
  if(s->state != P_NET_CREATED)
    {
    P_error_set_errno(P_EBADSTATE);
    return(FALSE);
    }
  
  if(!__P_socket_addr2sock(&(s->raddr), host))
    {
    P_error_set_errno(P_EADDRINFO);
    return(FALSE);
    }

  s->raddr.sin_port = htons(port);

CONNECT:
  if(connect(s->sd, (struct sockaddr *)&(s->raddr), sz) < 0)
    {
    switch(errno)
      {
      case ECONNREFUSED:
        P_error_set_errno(P_ENOCONN);
	break;
        
      case EINTR:
	goto CONNECT;
        
      default:
	P_error_set_errno(P_ECONNECT);
      }

    return(FALSE);
    }
  
  s->state = P_NET_CONNECTED;
  sz = (int)sizeof(struct sockaddr_in);
  getsockname(s->sd, (struct sockaddr *)&(s->laddr), &sz);

  return(TRUE);
  }

/*
 */

bool_t P_socket_shutdown(p_socket_t *s, uint_t how)
  {

  if(!s || (how < P_NET_SHUTRD) || (how > P_NET_SHUTALL))
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }
  
  if(s->state != P_NET_CONNECTED)
    {
    P_error_set_errno(P_EBADSTATE);
    return(FALSE);
    }

  s->state = P_NET_SHUTDOWN;
  s->flags |= (how << P_NET_OSHUT);
  shutdown(s->sd, --how);

  return(TRUE);
  }

/*
 */

bool_t P_socket_destroy(p_socket_t *s)
  {

  if(!s)
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }

  if(!((s->state == P_NET_CREATED) || ((s->state == P_NET_SHUTDOWN)
				       && (s->flags & P_NET_MSHUT))))
    {
    P_error_set_errno(P_EBADSTATE);
    return(FALSE);
    }
  
  if(s->sfp)
    fclose(s->sfp);

  close(s->sd);

  return(TRUE);
  }

/*
 */

bool_t P_socket_fopen(p_socket_t *s, int buffering)
  {

  if(!s)
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }
  
  if(s->state != P_NET_CONNECTED)
    {
    P_error_set_errno(P_EBADSTATE);
    return(FALSE);
    }
  
  if(s->type != P_NET_TCP)
    {
    P_error_set_errno(P_EBADTYPE);
    return(FALSE);
    }
  
  if(!(s->sfp = fdopen(s->sd, "r+")))
    {
    P_error_set_errno(P_EFDOPEN);
    return(FALSE);
    }

  if(setvbuf(s->sfp, NULL, buffering, 0))
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }
  
  return(TRUE);
  }

/*
 */

bool_t P_socket_get_peeraddr(p_socket_t *s, char *buf, size_t bufsz)
  {

  if(!s || !buf || !bufsz)
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }
  
  if(s->state != P_NET_CONNECTED)
    {
    P_error_set_errno(P_EBADSTATE);
    return(FALSE);
    }
  
  if(!__P_socket_sock2addr(&(s->raddr), buf, bufsz))
    {
    P_error_set_errno(P_EADDRINFO);
    return(FALSE);
    }
  
  return(TRUE);
  }

/*
 */

bool_t P_socket_fclose(p_socket_t *s)
  {

  if(!s)
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }

  if(s->sfp)
    {
    fclose(s->sfp);
    s->sfp = NULL;
    }
  
  return(TRUE);
  }

/*
 */

bool_t P_socket_reopen(int sd, p_socket_t *s)
  {
  int flags, i, type, sz;

  if(!s)
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }
  
  if((flags = fcntl(sd, F_GETFL, 0)) == -1)
    {
    P_error_set_errno(P_EFCNTL);
    return(FALSE);
    }

  /* get socket type */

  sz = (int)sizeof(int);
  if(getsockopt(sd, SOL_SOCKET, SO_TYPE, (char *)&type, &sz))
    {
    P_error_set_errno(P_ESOCKINFO);
    return(FALSE);
    }
  
  for(i = 0; i < P_NET_MAXTYPE; i++)
    if(type == __P_net_socktypes[i]) break;
  if(i == P_NET_MAXTYPE)
    {
    P_error_set_errno(P_EBADTYPE);
    return(FALSE);
    }

  s->sd = sd;
  s->type = i;
  s->state = P_NET_CONNECTED;

  /* get socket's local address */

  i = sizeof(struct sockaddr_in);
  if(getsockname(s->sd, (struct sockaddr *)&(s->laddr), &i) < 0)
    {
    P_error_set_errno(P_ESOCKINFO);
    return(FALSE);
    }

  /* get socket's remote address */

  i = sizeof(struct sockaddr_in);
  if(getpeername(sd, (struct sockaddr *)&(s->raddr), &i) < 0)
    {
    switch(errno)
      {
      case ENOTCONN:
        s->state = P_NET_CREATED;
        break;
        
      default:
	P_error_set_errno(P_ESOCKINFO);
	return(FALSE);
      }
    }

  /* get flags */

  if(flags & O_NONBLOCK)
    s->flags |= P_NET_MUNBLOCK;

  return(TRUE);
  }

/*
 */

bool_t P_socket_set_option(p_socket_t *s, uint_t option, bool_t flag,
                           uint_t value)
  {

  if(!s)
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }

  switch(option)
    {
    /* block */

    case P_NET_OPT_BLOCK:
      {
      int flags;

      if(s->state == P_NET_SHUTDOWN)
	{
	P_error_set_errno(P_EBADSTATE);
	return(FALSE);
	}
      
      if((flags = fcntl(s->sd, F_GETFL, 0)) == -1)
	{
	P_error_set_errno(P_EFCNTL);
	return(FALSE);
	}
      
      flag ? (flags &= ~O_NONBLOCK) : (flags |= O_NONBLOCK);
      flag ? (s->flags &= ~P_NET_MUNBLOCK) : (s->flags |= P_NET_MUNBLOCK);
      if(fcntl(s->sd, F_SETFL, flags) == -1)
	{
	P_error_set_errno(P_EFCNTL);
	return(FALSE);
	}

      return(TRUE);
      }

    /* linger */
      
    case P_NET_OPT_LINGER:
      {
      struct linger l;

      l.l_onoff = (flag ? 1 : 0);
      l.l_linger = value;

      if(setsockopt(s->sd, SOL_SOCKET, SO_LINGER, (char *)&l,
		    sizeof(struct linger)) != 0)
	{
	P_error_set_errno(P_ESOCKINFO);
	return(FALSE);
	}
      return(TRUE);
      }

    /* reuse address */
      
    case P_NET_OPT_REUSEADDR:
      {
      int v = flag;
      
      if(setsockopt(s->sd, SOL_SOCKET, SO_REUSEADDR, (char *)&v, sizeof(int))
         != 0)
        {
	P_error_set_errno(P_ESOCKINFO);
        return(FALSE);
        }
      
      return(TRUE);
      }

      /* oob data inline */

    case P_NET_OPT_OOBINLINE:
      {
      int v = flag;

      if(setsockopt(s->sd, SOL_SOCKET, SO_OOBINLINE, (char *)&v, sizeof(int))
         != 0)
        {
	P_error_set_errno(P_ESOCKINFO);
        return(FALSE);
        }
      
      return(TRUE);      
      }

    /* keepalive */

    case P_NET_OPT_KEEPALIVE:
      {
      int v = flag;

      if(setsockopt(s->sd, SOL_SOCKET, SO_KEEPALIVE, (char *)&v, sizeof(int))
         != 0)
        {
	P_error_set_errno(P_ESOCKINFO);
        return(FALSE);
        }
      
      return(TRUE);      
      }

    /* recv buffer */

    case P_NET_OPT_RECVBUF:
      {
      int v = (int)value;

      if(setsockopt(s->sd, SOL_SOCKET, SO_RCVBUF, (char *)&v, sizeof(int))
         != 0)
        {
	P_error_set_errno(P_ESOCKINFO);
        return(FALSE);
        }
      
      return(TRUE);      
      }

    /* send buffer */

    case P_NET_OPT_SENDBUF:
      {
      int v = (int)value;

      if(setsockopt(s->sd, SOL_SOCKET, SO_SNDBUF, (char *)&v, sizeof(int))
         != 0)
        {
	P_error_set_errno(P_ESOCKINFO);
        return(FALSE);
        }
      
      return(TRUE);      
      }

    /* unknown*/
      
    default:
      P_error_set_errno(P_EINVAL);
      return(FALSE);
    }
  }

/*
 */

extern bool_t P_socket_get_option(p_socket_t *s, uint_t option, bool_t *flag,
                                  uint_t *value)
  {
  if(!s || !flag || !value)
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }

  switch(option)
    {
    /* block */

    case P_NET_OPT_BLOCK:
      {
      return(!(s->flags & P_NET_MUNBLOCK));
      }

    /* linger*/

    case P_NET_OPT_LINGER:
      {
      int sz = sizeof(struct linger);
      struct linger l;

      if(getsockopt(s->sd, SOL_SOCKET, SO_LINGER, (char *)&l, &sz) != 0)
        {
	P_error_set_errno(P_ESOCKINFO);
        return(FALSE);        
        }

      *flag = l.l_onoff ? TRUE : FALSE;
      *value = (uint_t)l.l_linger;

      return(TRUE);
      }

    /* reuse address*/
      
    case P_NET_OPT_REUSEADDR:
      {
      int sz = sizeof(int);
      int v;
      
      if(getsockopt(s->sd, SOL_SOCKET, SO_REUSEADDR, (char *)&v, &sz) != 0)
        {
	P_error_set_errno(P_ESOCKINFO);
        return(FALSE);        
        }      

      *flag = v ? TRUE : FALSE;
      
      return(TRUE);
      }

    /* oob data inline*/
      
    case P_NET_OPT_OOBINLINE:
      {
      int sz = sizeof(int);
      int v;
      
      if(getsockopt(s->sd, SOL_SOCKET, SO_OOBINLINE, (char *)&v, &sz) != 0)
        {
	P_error_set_errno(P_ESOCKINFO);
        return(FALSE);        
        }      

      *flag = v ? TRUE : FALSE;
      
      return(TRUE);
      }

    /* keepalive*/

    case P_NET_OPT_KEEPALIVE:
      {
      int sz = sizeof(int);
      int v;
      
      if(getsockopt(s->sd, SOL_SOCKET, SO_KEEPALIVE, (char *)&v, &sz) != 0)
        {
	P_error_set_errno(P_ESOCKINFO);
        return(FALSE);        
        }      

      *flag = v ? TRUE : FALSE;
      
      return(TRUE);
      }

    /* recv buffer*/

    case P_NET_OPT_RECVBUF:
      {
      int sz = sizeof(int);
      int v;

      if(getsockopt(s->sd, SOL_SOCKET, SO_RCVBUF, (char *)&v, &sz) != 0)
        {
	P_error_set_errno(P_ESOCKINFO);
        return(FALSE);
        }

      *value = (uint_t)v;
      
      return(TRUE);      
      }

    /* send buffer*/

    case P_NET_OPT_SENDBUF:
      {
      int sz = sizeof(int);
      int v;

      if(getsockopt(s->sd, SOL_SOCKET, SO_SNDBUF, (char *)&v, &sz) != 0)
        {
	P_error_set_errno(P_ESOCKINFO);
        return(FALSE);
        }

      *value = (uint_t)v;
      
      return(TRUE);      
      }

    /* unknown*/

    default:
      P_error_set_errno(P_EINVAL);
      return(FALSE);
    }
  }

/*
 */

bool_t P_socket_mcast_join(p_socket_t *s, const char *addr)
  {
  struct ip_mreq mreq;
  struct sockaddr_in saddr;

  if(! s || !addr)
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }

  if(!*addr)
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }

  if(s->type != P_NET_UDP)
    {
    P_error_set_errno(P_EBADTYPE);
    return(FALSE);
    }
  
  if(! __P_socket_addr2sock(&saddr, addr))
    {
    P_error_set_errno(P_EADDRINFO);
    return(FALSE);
    }
  
  memcpy(&mreq.imr_multiaddr, &(saddr.sin_addr), sizeof(struct in_addr));
  mreq.imr_interface.s_addr = htonl(INADDR_ANY);

  if(setsockopt(s->sd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0)
    {
    P_error_set_errno(P_ESOCKINFO);
    return(FALSE);
    }
  
  return(TRUE);
  }

/*
 */

bool_t P_socket_mcast_leave(p_socket_t *s, const char *addr)
  {
  struct ip_mreq mreq;
  struct sockaddr_in saddr;

  if(! s || !addr)
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }

  if(!*addr)
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }

  if(s->type != P_NET_UDP)
    {
    P_error_set_errno(P_EBADTYPE);
    return(FALSE);
    }
  
  if(! __P_socket_addr2sock(&saddr, addr))
    {
    P_error_set_errno(P_EADDRINFO);
    return(FALSE);
    }
  
  memcpy(&mreq.imr_multiaddr, &(saddr.sin_addr), sizeof(struct in_addr));
  mreq.imr_interface.s_addr = htonl(INADDR_ANY);

  if(setsockopt(s->sd, IPPROTO_IP, IP_DROP_MEMBERSHIP, &mreq, sizeof(mreq))
     < 0)
    {
    P_error_set_errno(P_ESOCKINFO);
    return(FALSE);
    }
  
  return(TRUE);
  }

/*
 */

bool_t P_socket_mcast_set_ttl(p_socket_t *s, byte_t ttl)
  {
  if(!s)
    {
    P_error_set_errno(P_EINVAL);
    return(FALSE);
    }

  if(s->type != P_NET_UDP)
    {
    P_error_set_errno(P_EBADTYPE);
    return(FALSE);
    }

  if(setsockopt(s->sd, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl)) < 0)
    {
    P_error_set_errno(P_ESOCKINFO);
    return(FALSE);
    }

  return(TRUE);
  }

/* end of source file */
