/* ----------------------------------------------------------------------------
   Ping Utility Library
   Copyright (C) 1994-2000 PING Software Group and Mark A Lindner

   This file is part of pingutil.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   ----------------------------------------------------------------------------
   $Log: sockio.c,v $
   Revision 1.2  2000/07/04 12:54:39  markl
   More fixes.

   Revision 1.1  2000/06/14 06:02:48  ping
   Initial checkin.

   ----------------------------------------------------------------------------
*/

/* Feature test switches */

#include "config.h"

/* Local headers */

#include "pingutil.h"
#include "pingnet.h"
#include "perrno.h"
#include "netcommon.h"

/* Functions */

int P_socket_send(p_socket_t *s, const char *buf, size_t bufsz, bool_t oobf)
  {
  int b;

  if(!s || !buf || !bufsz)
    {
    P_error_set_errno(P_EINVAL);
    return(-1);
    }

  if(s->state != P_NET_CONNECTED)
    {
    P_error_set_errno(P_EBADSTATE);
    return(-1);
    }

  switch(s->type)
    {
    case P_NET_TCP:
      {
      int flags = (oobf ? MSG_OOB : 0), bsofar = 0, bleft = (int)bufsz;
      char *p = (char *)buf;

      do
	{
SEND1:	
	b = send(s->sd, p, bleft, flags);
        
	if(b == 0)
	  {
	  P_socket_shutdown(s, P_NET_SHUTALL);
	  P_error_set_errno(P_ELOSTCONN);
	  return(-bsofar);
	  }
        
	else if(b < 0)
	  {
	  switch(errno)
	    {
	    case EWOULDBLOCK:
	      P_error_set_errno(P_EBLOCKED);
	      return(bsofar);
              
	    case EINTR:
	      goto SEND1;
              
	    default:
	      P_error_set_errno(P_ESEND);
	      return(-bsofar);
	    }
	  }
        
	else bleft -= b, bsofar += b, p += b;
	}
      while(bleft);

      return(bsofar);
      }

    case P_NET_UDP:
      {      
SENDTO1:      
      b = sendto(s->sd, buf, (int)bufsz, 0,
                 ((s->state == P_NET_CONNECTED)
                 ? NULL : (struct sockaddr *)&(s->raddr)),
                 ((s->state == P_NET_CONNECTED)
                 ? 0 : (int)sizeof(struct sockaddr_in)));

      if(b == 0)
	{
	P_socket_shutdown(s, P_NET_SHUTALL);
	P_error_set_errno(P_ELOSTCONN);
	return(0);
	}
      
      else if(b < 0)
	{
	switch(errno)
	  {
	  case EMSGSIZE:
	    P_error_set_errno(P_EMSG2BIG);
	    return(-1);
            
	  case EWOULDBLOCK:
	    P_error_set_errno(P_EBLOCKED);
	    return(0);
            
	  case EINTR:
	    goto SENDTO1;
            
	  default:
	    P_error_set_errno(P_ESEND);
	    return(-1);
          }
	}
      
      else return(b);
      }

    default:
      P_error_set_errno(P_EBADTYPE);
      return(-1);
    }
  }

/*
 */

int P_socket_recv(p_socket_t *s, char *buf, size_t bufsz, bool_t oobf)
  {
  int b;

  if(!s || !buf || !bufsz)
    {
    P_error_set_errno(P_EINVAL);
    return(-1);
    }

  if(s->state != P_NET_CONNECTED)
    {
    P_error_set_errno(P_EBADSTATE);
    return(-1);
    }

  switch(s->type)
    {
    case P_NET_TCP:
      {
      int flags = (oobf ? MSG_OOB : 0), bsofar = 0, bleft = (int)bufsz;
      char *p = buf;

      do
	{
RECV1:
	b = recv(s->sd, p, bleft, flags);

	if(b == 0)
	  {
	  P_socket_shutdown(s, P_NET_SHUTALL);
	  P_error_set_errno(P_ELOSTCONN);
	  return(-bsofar);
	  }

	else if(b < 0)
	  {
	  switch(errno)
	    {
	    case EWOULDBLOCK:
	      P_error_set_errno(P_EBLOCKED);
	      return(bsofar);
              
	    case EINTR:
	      goto RECV1;
              
	    default:
	      P_error_set_errno(P_ERECV);
	      return(-bsofar);
	    }
	  }
        
	else bleft -= b, bsofar += b, p += b;
	}
      while(bleft);
      
      return(bsofar);
      }

    case P_NET_UDP:
      {
      int sz = (int)sizeof(struct sockaddr_in);

RECVFROM1:
      b = recvfrom(s->sd, buf, (int)bufsz, 0,
                   ((s->state == P_NET_CONNECTED)
                    ? NULL : (struct sockaddr *)&(s->raddr)),
                   ((s->state == P_NET_CONNECTED)
                    ? NULL : &sz));

      if(b == 0)
	{
	P_socket_shutdown(s, P_NET_SHUTALL);
	P_error_set_errno(P_ELOSTCONN);
	return(0);
	}

      else if(b < 0)
	{
	switch(errno)
	  {
	  case EMSGSIZE:
	    P_error_set_errno(P_EMSG2BIG);
	    return(-1);
            
	  case EWOULDBLOCK:
	    P_error_set_errno(P_EBLOCKED);
	    return(0);
            
	  case EINTR:
	    goto RECVFROM1;
            
	  default:
	    P_error_set_errno(P_ERECV);
	    return(-1);
	  }
	}
      
      else return(b);
      }

    default:
      P_error_set_errno(P_EBADTYPE);
      return(-1);
    }
  }

/*
 */

int P_socket_sendto(p_socket_t *s, const char *buf, size_t bufsz,
		    const char *addr, in_port_t port)
  {
  int b;

  if(!s || !buf || !bufsz)
    {
    P_error_set_errno(P_EINVAL);
    return(-1);
    }
  
  if(s->state != P_NET_CREATED)
    {
    P_error_set_errno(P_EBADSTATE);
    return(-1);
    }

  if(s->type != P_NET_UDP)
    {
    P_error_set_errno(P_EBADTYPE);
    return(-1);
    }

  if(addr)
    {
    if(! *addr)
      {
      P_error_set_errno(P_EINVAL);
      return(-1);
      }
    
    if(!__P_socket_addr2sock(&(s->raddr), addr))
      {
      P_error_set_errno(P_EADDRINFO);
      return(-1);
      }
    
    s->raddr.sin_port = htons(port);
    }
  
SENDTO2:
  b = sendto(s->sd, buf, (int)bufsz, 0, (struct sockaddr *)&(s->raddr),
	     (int)sizeof(struct sockaddr_in));

  if(b == 0)
    {
    P_socket_shutdown(s, P_NET_SHUTALL);
    P_error_set_errno(P_ELOSTCONN);
    return(-1);
    }
  
  else if(b < 0)
    {
    switch(errno)
      {
      case EMSGSIZE:
        P_error_set_errno(P_EMSG2BIG);
        return(-1);

      case EWOULDBLOCK:
	P_error_set_errno(P_EBLOCKED);
	return(0);
        
      case EINTR:
	goto SENDTO2;
        
      default:
	P_error_set_errno(P_ESENDTO);
	return(-1);
      }
    }
  
  else return(b);
  }

/*
 */

int P_socket_sendreply(p_socket_t *s, const char *buf, size_t bufsz)
  {
  int b;

  if(!s || !buf || !bufsz)
    {
    P_error_set_errno(P_EINVAL);
    return(-1);
    }
  
  if(s->state != P_NET_CREATED)
    {
    P_error_set_errno(P_EBADSTATE);
    return(-1);
    }
  
  if(s->type != P_NET_UDP)
    {
    P_error_set_errno(P_EBADTYPE);
    return(-1);
    }

SENDTO3:
  b = sendto(s->sd, buf, (int)bufsz, 0, (struct sockaddr *)&(s->raddr),
	     (int)sizeof(struct sockaddr_in));

  if(b == 0)
    {
    P_socket_shutdown(s, P_NET_SHUTALL);
    P_error_set_errno(P_ELOSTCONN);
    return(-1);
    }

  else if(b < 0)
    {
    switch(errno)
      {
      case EMSGSIZE:
        P_error_set_errno(P_EMSG2BIG);
        return(-1);
        
      case EWOULDBLOCK:
	P_error_set_errno(P_EBLOCKED);
	return(0);
        
      case EINTR:
	goto SENDTO3;
        
      default:
	P_error_set_errno(P_ESENDTO);
	return(-1);
      }
    }
  
  else return(b);
  }

/*
 */

int P_socket_recvfrom(p_socket_t *s, char *buf, size_t bufsz, char *addr,
		      size_t addrsz)
  {
  int b, sz = (int)sizeof(struct sockaddr_in);

  if(!s || !buf || !bufsz)
    {
    P_error_set_errno(P_EINVAL);
    return(-1);
    }
  
  if(s->state != P_NET_CREATED)
    {
    P_error_set_errno(P_EBADSTATE);
    return(-1);
    }
  
  if(s->type != P_NET_UDP)
    {
    P_error_set_errno(P_EBADTYPE);
    return(-1);
    }

RECVFROM2:
  b = recvfrom(s->sd, buf, (int)bufsz, 0, (struct sockaddr *)&(s->raddr), &sz);

  if(b == 0)
    {
    P_socket_shutdown(s, P_NET_SHUTALL);
    P_error_set_errno(P_ELOSTCONN);
    return(-1);
    }
  
  else if(b < 0)
    {
    switch(errno)
      {
      case EMSGSIZE:
        P_error_set_errno(P_EMSG2BIG);
        return(-1);
        
      case EWOULDBLOCK:
	P_error_set_errno(P_EBLOCKED);
	return(0);
        
      case EINTR:
	goto RECVFROM2;
        
      default:
	P_error_set_errno(P_ERECVFROM);
	return(-1);
      }
    }
  else
    {
    if(addr)
      {
      if(!addrsz)
	{
	P_error_set_errno(P_EINVAL);
	return(-1);
	}

      __P_socket_sock2addr(&(s->raddr), addr, addrsz);
      }
    
    return(b);
    }
  }

/*
 */

int P_socket_recvreply(p_socket_t *s, char *buf, size_t bufsz)
  {
  int b, sz = (int)sizeof(struct sockaddr_in);

  if(!s || !buf || !bufsz)
    {
    P_error_set_errno(P_EINVAL);
    return(-1);
    }
  
  if(s->state != P_NET_CREATED)
    {
    P_error_set_errno(P_EBADSTATE);
    return(-1);
    }
  
  if(s->type != P_NET_UDP)
    {
    P_error_set_errno(P_EBADTYPE);
    return(-1);
    }

RECVFROM3:
  b = recvfrom(s->sd, buf, (int)bufsz, 0, (struct sockaddr *)&(s->raddr), &sz);

  if(b == 0)
    {
    P_socket_shutdown(s, P_NET_SHUTALL);
    P_error_set_errno(P_ELOSTCONN);
    return(-1);
    }

  else if(b < 0)
    {
    switch(errno)
      {
      case EMSGSIZE:
        P_error_set_errno(P_EMSG2BIG);
        return(-1);
        
      case EWOULDBLOCK:
	P_error_set_errno(P_EBLOCKED);
	return(0);
        
      case EINTR:
	goto RECVFROM3;
        
      default:
	P_error_set_errno(P_ERECVFROM);
	return(-1);
      }
    }

  else return(b);
  }

/*
 */

static int __P_socket_write(int sd, const char *buf, size_t bufsz,
			    uint_t slen, uint_t *snum)
  {
  char *p = (char *)buf;
  int bsofar = 0, bleft = (int)bufsz, b, e;
  static struct timeval tv;
  fd_set fdset;

  tv.tv_usec = 0, tv.tv_sec = slen;
  FD_ZERO(&fdset);
  FD_SET(sd, &fdset);

  for(;;)
    {
SELECT:
    e = select(FD_SETSIZE, NULL, &fdset, NULL, &tv);

    if(e == 0)
      {
      P_error_set_errno(P_ETIMEOUT);
      return(-bsofar);
      }
    
    else if(e < 0)
      {
      switch(errno)
	{
	case EINTR:
	  goto SELECT;
          
	default:
	  P_error_set_errno(P_ESELECT);
	  return(-bsofar);
	}
      }

SEND3:      
    b = send(sd, p, bleft, 0);
    
    if(b == 0)
      {
      P_error_set_errno(P_ELOSTCONN);
      return(-bsofar);
      }
    
    else if(b < 0)
      {
      switch(errno)
	{
	case EWOULDBLOCK:
	  if(!(*snum)--)
	    {
	    P_error_set_errno(P_ETIMEOUT);
	    return(-bsofar);
	    }
	  else continue;
	  break;
          
	case EINTR:
	  goto SEND3;
          
	default:
	  P_error_set_errno(P_ESEND);
	  return(-bsofar);
	  }
      }
    else
      {
      p += b, bsofar += b;
      if(!(bleft -= b))
        break;
      }
    }
  
  return(bsofar);
  }

/*
 */

static int __P_socket_read(int sd, char *buf, size_t bufsz, char termin,
			   uint_t slen, uint_t *snum)
  {
  char *p = buf, *q;
  int bsofar = 0, bleft = (int)(--bufsz), i, b, e;
  static struct timeval tv;
  fd_set fdset;

  tv.tv_usec = 0, tv.tv_sec = slen;
  FD_ZERO(&fdset);
  FD_SET(sd, &fdset);

  for(;;)
    {

SELECT2:
    e = select(FD_SETSIZE, &fdset, NULL, NULL, &tv);

    if(e == 0)
      {
      P_error_set_errno(P_ETIMEOUT);
      return(-bsofar);
      }
    
    else if(e < 0)
      {
      switch(errno)
	{
	case EINTR:
	  goto SELECT2;

	default:
	  P_error_set_errno(P_ESELECT);
	  return(-bsofar);
	}
      }

RECV3:
    b = recv(sd, p, bleft, MSG_PEEK);

    if(b == 0)
      {
      P_error_set_errno(P_ELOSTCONN);
      return(-bsofar);
      }
    
    else if(b < 0)
      {
      switch(errno)
	{
	case EWOULDBLOCK:
	  if(!(*snum)--)
	    {
	    P_error_set_errno(P_ETIMEOUT);
	    return(-bsofar);
	    }
	  else continue;
	  break;
          
	case EINTR:
	  goto RECV3;
          
	default:
	  P_error_set_errno(P_ERECV);
	  return(-bsofar);
	  }
      }

    else
      {
      /* try to find terminator */

      for(q = p, i = 0; i < b; q++, i++)
        if(*q == termin)
	  {
          b = bleft = i + 1;
          break;
	  }

RECV4:
      if((i = recv(sd, p, b, 0)) != b)
	{
	if((i < 0) && (errno == EINTR))
          goto RECV4;

	P_error_set_errno(P_ERECV);
	return(-bsofar);
	}
      
      bsofar += b;
      if(!(bleft -= b)) break;
      p += b;
      }
    }
  *(buf + bsofar) = NUL;
  
  return(bsofar);
  }

/*
 */

int P_socket_writeline(p_socket_t *s, const char *buf, const char *termin,
		       uint_t slen, uint_t snum)
  {
  int b, c;

  if(!s || (!buf && !termin))
    {
    P_error_set_errno(P_EINVAL);
    return(-1);
    }
  
  if(!((s->state == P_NET_CONNECTED)
     && !(s->flags & P_NET_MUNBLOCK)
     && !P_bit_isset(s->flags, P_NET_OSHUTWR)))
    {
    P_error_set_errno(P_EBADSTATE);
    return(-1);
    }
  
  if(s->type != P_NET_TCP)
    {
    P_error_set_errno(P_EBADTYPE);
    return(-1);
    }

  if(buf)
    if(*buf)
      if((b = __P_socket_write(s->sd, buf, strlen(buf), slen, &snum)) <= 0)
        return(b);

  if(termin)
    if(*termin)
      {
      if((c = __P_socket_write(s->sd, termin, strlen(termin), slen, &snum))
	 <= 0)
      return(-c + b);
      else b += c;
      }
  
  return(b);
  }

/*
 */

int P_socket_readline(p_socket_t *s, char *buf, size_t bufsz, char termin,
		      uint_t slen, uint_t snum)
  {
  
  if(!s || !buf || !bufsz)
    {
    P_error_set_errno(P_EINVAL);
    return(-1);
    }
  
  if(!((s->state == P_NET_CONNECTED)
       && !(s->flags & P_NET_MUNBLOCK)
       && !P_bit_isset(s->flags, P_NET_OSHUTRD)))
    {
    P_error_set_errno(P_EBADSTATE);
    return(-1);
    }
  
  if(s->type != P_NET_TCP)
    {
    P_error_set_errno(P_EBADTYPE);
    return(-1);
    }

  return(__P_socket_read(s->sd, buf, bufsz, termin, slen, &snum));
  }

/* end of source file */
