/* ----------------------------------------------------------------------------
   Ping Utility Library
   Copyright (C) 1994-2000 PING Software Group and Mark A Lindner

   This file is part of pingutil.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   ----------------------------------------------------------------------------
   $Log: strheap.c,v $
   Revision 1.2  2000/07/04 12:54:39  markl
   More fixes.

   Revision 1.1  2000/06/14 06:02:48  ping
   Initial checkin.

   ----------------------------------------------------------------------------
*/

/* Feature test switches */

#include "config.h"

/* Local headers */

#include "pingutil.h"

/* Macros */

#define P_STRHEAP_BLOCKSZ 40
#define P_STRHEAP_IBLOCKSZ 40
#define P_STRHEAP_DELETED 1
#define P_STRHEAP_SORTED 2
#define P_STRHEAP_FRAG 4

/* Functions */

p_strheap_t *P_strheap_create(uint_t blocksz)
  {
  p_strheap_t *h;

  if(blocksz < P_STRHEAP_BLOCKSZ)
    return(NULL);

  h = P_new(p_strheap_t);
  h->mem = P_newstr(blocksz);
  h->index = P_malloc(P_STRHEAP_IBLOCKSZ, struct p_strheap_ent);
  h->blk = h->iblk = 1;
  h->bot = h->flags = h->size = 0;
  h->blocksz = blocksz;

  return(h);
  }

/*
 */

void P_strheap_destroy(p_strheap_t *h)
  {
  if(!h)
    return;

  P_free(h->mem);
  P_free(h->index);
  P_free(h);
  }

/*
 */

uint_t P_strheap_store(p_strheap_t *h, const char *s)
  {
  size_t len;
  uint_t freespace;
  struct p_strheap_ent *n;

  if(!h || !s)
    return(0);

  len = strlen(s) + 1;
  freespace = (h->blk * h->blocksz) - h->bot;
  if(len >= freespace)
    {
    h->blk += (((len - freespace) / h->blocksz) + 1);
    h->mem = P_realloc(h->mem, h->blk * h->blocksz, char);
    }

  memcpy((void *)(h->mem + h->bot), (void *)s, len);
  if(!(h->size) % P_STRHEAP_IBLOCKSZ)
    h->index = P_realloc(h->index, ++(h->iblk) * P_STRHEAP_IBLOCKSZ,
                         struct p_strheap_ent);

  n = &(h->index[h->size]);
  n->offset = h->bot;
  n->len = len;
  h->bot += len;

  return(h->size + 1);
  }

/*
 */

char *P_strheap_restore(p_strheap_t *h, uint_t index)
  {
  struct p_strheap_ent *p;

  if(!h || !index || index > h->size)
    return(NULL);

  p = &(h->index[--index]);
  return((p->flags & P_STRHEAP_DELETED) ? NULL : (h->mem + p->offset));
  }

/*
 */

bool_t P_strheap_delete(p_strheap_t *h, uint_t index)
  {
  struct p_strheap_ent *p;

  if(!h || !index)
    return(FALSE);
  if(index > h->size)
    return(FALSE);

  p = &(h->index[--index]);
  if(p->flags & P_STRHEAP_DELETED)
    return(FALSE);

  p->flags |= P_STRHEAP_DELETED;
  h->flags |= P_STRHEAP_FRAG;

  return(TRUE);
  }

/*
 */

p_strheap_t *P_strheap_load(const char *path)
  {
  FILE *fp;
  p_strheap_t *h;

  if(!path)
    return(NULL);
  if(!*path)
    return(NULL);
  if(!(fp = fopen(path, "r")))
    return(NULL);

  h = P_new(p_strheap_t);
  if((fread((void *)h, sizeof(p_strheap_t), (size_t)1, fp)) != 1)
    {
    P_free(h);
    fclose(fp);
    return(NULL);
    }

  h->index = P_newa(h->iblk * P_STRHEAP_IBLOCKSZ, struct p_strheap_ent);
  if((fread((void *)h->index, sizeof(struct p_strheap_ent),
	    (size_t)(h->iblk * P_STRHEAP_IBLOCKSZ), fp))
     != (size_t)(h->iblk * P_STRHEAP_IBLOCKSZ))
    {
    P_free(h->index);
    P_free(h);
    fclose(fp);
    return(NULL);
    }

  h->mem = P_newstr(h->blk * P_STRHEAP_BLOCKSZ);
  if((fread((void *)h->mem, sizeof(char), (size_t)(h->blk * h->blocksz), fp))
     != (size_t)(h->blk * h->blocksz))
    {
    P_free(h->mem);
    P_free(h->index);
    P_free(h);
    fclose(fp);
    return(NULL);
    }
  fclose(fp);

  return(NULL);
  }

/*
 */

bool_t P_strheap_save(p_strheap_t *h, const char *path)
  {
  FILE *fp;

  if(!path || !h)
    return(FALSE);
  if(!*path)
    return(FALSE);
  if(!(fp = fopen(path, "w")))
    return(FALSE);

  if((fwrite((void *)h, sizeof(p_strheap_t), (size_t)1, fp)) != 1)
    {
    fclose(fp);
    return(FALSE);
    }

  if((fwrite((void *)h->index, sizeof(struct p_strheap_ent),
	     (size_t)(h->iblk * P_STRHEAP_IBLOCKSZ), fp))
     != (size_t)(h->iblk * P_STRHEAP_IBLOCKSZ))
    {
    fclose(fp);
    return(FALSE);
    }

  if((fwrite((void *)h->mem, sizeof(char), (size_t)(h->blk * h->blocksz), fp))
     != (size_t)(h->blk * h->blocksz))
    {
    fclose(fp);
    return(FALSE);
    }

  fclose(fp);
  return(TRUE);
  }

/* end of source file */
