/* ----------------------------------------------------------------------------
   Ping Utility Library
   Copyright (C) 1994-2000 PING Software Group and Mark A Lindner

   This file is part of pingutil.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   ----------------------------------------------------------------------------
   $Log: system.c,v $
   Revision 1.10  2000/12/31 03:57:25  markl
   Fixed goto label problem.

   Revision 1.9  2000/09/23 09:28:27  markl
   Added ctermid_r() check.

   Revision 1.8  2000/08/17 07:23:31  markl
   Added new macro 'THREADED_LIBRARY' to correct some flaws in the
   configuration logic.

   Revision 1.7  2000/08/17 06:37:45  markl
   Changes to eliminate most compiler warnings.

   Revision 1.6  2000/08/13 05:56:43  markl
   More Linux ports.

   Revision 1.5  2000/07/21 08:14:17  markl
   Added new password functions.

   Revision 1.4  2000/07/04 12:54:39  markl
   More fixes.

   Revision 1.3  2000/06/14 21:03:07  markl
   Reentrancy fixes.

   Revision 1.2  2000/06/14 06:34:15  markl
   Removed feature test switch.

   Revision 1.1  2000/06/14 06:02:49  ping
   Initial checkin.

   ----------------------------------------------------------------------------
*/

/* Feature test switches */

#include "config.h"

/* System headers */

#include <sys/utsname.h>
#include <grp.h>
#include <pwd.h>
#include <crypt.h>

/* Local headers */

#include "pingutil.h"

/* Macros */

#define P_SYSTEM_SALT_STRLEN 64

/* File scope variables */

static const char *__P_system_salt_string = "abcdefghijklmnopqrstuvwxyz" \
  "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";

static p_sysinfo_t *__P_system_info = NULL;

/* Functions */

bool_t P_system_isingroup(const char *login, const char *group)
  {
  bool_t foundf = FALSE;
  static struct group *gr = NULL;
  struct group *rgr;
  char **u, *buf;
  int blk = 128;
#if defined(_POSIX_PTHREAD_SEMANTICS) || defined (__hpux) || defined (linux)
  int x;
#endif

  if(!login || !group)
    return(FALSE);
  if(!*login || !*group)
    return(FALSE);

  if(!gr)
    gr = P_new(struct group);


#if defined(_POSIX_PTHREAD_SEMANTICS) || (defined linux) || defined (__hpux) 
  
GETGRNAM:
  x = getgrnam_r(group, gr, buf, blk, &rgr);
  if(x == ERANGE)
    {
    blk += 128;
    buf = P_realloc(buf, blk, char);
    goto GETGRNAM;
    }
  else if(x != 0)
    return(FALSE);

#else

GETGRNAM:
  rgr = getgrnam_r(buf, gr, buf, blk);
  if(rgr == NULL)
    {
    if(errno == ERANGE)
      {
      buf = P_realloc(buf, blk, char);
      goto GETGRNAM;
      blk += 128;
      }
    else
      return(FALSE);
    }
  
#endif
  
  for(u = gr->gr_mem; *u; u++)
    {
    if(!strcmp(*u, login))
      {
      foundf = TRUE;
      break;
      }
    }

  P_free(buf);
  
  return(foundf);
  }

/*
 */

p_sysinfo_t *P_system_getinfo(void)
  {
  static struct passwd *pw = NULL;
  struct passwd *rpw;
  struct utsname uts;
  char *buf;
  int blk = 128;
#if defined(_POSIX_PTHREAD_SEMANTICS) || (defined linux) || defined (__hpux) 
  int x;
#endif

  if(!pw)
    pw = P_new(struct passwd);
  
  if(!__P_system_info)
    {
    __P_system_info = P_new(p_sysinfo_t);

    buf = P_newstr(blk);

    __P_system_info->uid = getuid();

GETPWUID:

#if defined(_POSIX_PTHREAD_SEMANTICS) || defined(linux) || defined (__hpux) 
    
    x = getpwuid_r(__P_system_info->uid, pw, buf, blk, &rpw);

    if(x == ERANGE)
      {
      blk += 128;
      buf = P_realloc(buf, blk, char);
      goto GETPWUID;
      }

#else

    rpw = getpwuid_r(__P_system_info->uid, pw, buf, blk);
    if(rpw == NULL)
      {
      if(errno == ERANGE)
        {
        blk += 128;
        buf = P_realloc(buf, blk, char);
        goto GETPWUID;
        }
      else
        pw = NULL;
      }    
    
#endif

    if(pw)
      {
      __P_system_info->login = P_string_dup(pw->pw_name);
      __P_system_info->fullname = P_string_dup(pw->pw_gecos);
      __P_system_info->homedir = P_string_dup(pw->pw_dir);
      __P_system_info->shell = P_string_dup(pw->pw_shell);
      }
    else
      {
      __P_system_info->login = NULL;
      __P_system_info->fullname = NULL;
      __P_system_info->homedir = NULL;
      __P_system_info->shell = NULL;
      }

    P_free(buf);
    
    __P_system_info->gid = getgid();
    __P_system_info->euid = geteuid();
    __P_system_info->egid = getegid();

    uname(&uts);
    __P_system_info->hostname = P_string_dup(uts.nodename);
    __P_system_info->osname = P_string_dup(uts.sysname);
    __P_system_info->osver = P_string_dup(uts.version);
    __P_system_info->osrel = P_string_dup(uts.release);
    __P_system_info->arch = P_string_dup(uts.machine);
    
    __P_system_info->pid = getpid();
    __P_system_info->ppid = getppid();
    __P_system_info->stime = time(NULL);
    __P_system_info->term = P_newstr(L_ctermid + 1);
#ifdef HAVE_CTERMID_R
    ctermid_r(__P_system_info->term);
#else
    ctermid(__P_system_info->term);
#endif
    __P_system_info->term[L_ctermid] = NUL;
    }

  return(__P_system_info);
  }

/*
 */

bool_t P_system_cdhome(void)
  {
  char *p = P_system_get_homedir();
  bool_t r = (chdir(p) ? FALSE : TRUE);

  P_free(p);
  return(r);
  }

/*
 */

uid_t P_system_get_uid(void)
  {
  return(P_system_getinfo()->uid);
  }

/*
 */

gid_t P_system_get_gid(void)
  {
  return(P_system_getinfo()->gid);
  }

/*
 */

pid_t P_system_get_pid(void)
  {
  return(P_system_getinfo()->pid);
  }

/*
 */

char *P_system_get_login(void)
  {
  return(P_system_getinfo()->login);
  }

/*
 */

char *P_system_get_fullname(void)
  {
  return(P_system_getinfo()->fullname);
  }

/*
 */

char *P_system_get_homedir(void)
  {
  return(P_system_getinfo()->homedir);
  }

/*
 */

char *P_system_get_hostname(void)
  {
  return(P_system_getinfo()->hostname);
  }

/*
 */

char *P_system_get_term(void)
  {
  return(P_system_getinfo()->term);
  }

/*
 */

bool_t P_system_passwd_validate(const char *plaintext, const char *ciphertext)
  {
  char salt[2], *c;

  strncpy(salt, ciphertext, 2);

  c = (char *)crypt(plaintext, salt);
  
  return(!strcmp(c, ciphertext));
  }

/*
 */

bool_t P_system_passwd_generate(const char *plaintext, char *buf,
                                size_t bufsz)
  {
  char salt[2], *cp;

  if(!plaintext || !buf || !bufsz)
    return(FALSE);
  
  salt[0] = __P_system_salt_string[P_random(P_SYSTEM_SALT_STRLEN)];
  salt[1] = __P_system_salt_string[P_random(P_SYSTEM_SALT_STRLEN)];
 
  cp = (char *)crypt(plaintext, salt);

  strncpy(buf, cp, bufsz - 1);
  *(buf + bufsz - 1) = NUL;

  return(TRUE);
  }

/* end of source file */

